# Backend selection

This directory contains files used to benchmark backend selection. The
goal of this workshop is to optimize backend selection made by
`use_backend` rules.

## Description

This is a simple example showing how to route traffic to different
backends based on the `host` header. This is typically the case when
you only have 1 single IP address with several services behind.

For the sake of this example the default backend is `default`, this is
where a request is routed if the hostname didn't match any of
configured values. If the hostname matches a known one the request is
routed to the `vhost` backend. In real life `vhost` would be several
different backends but to keep that example clean we only use 1
backend.

## Performance measurements

The `original` frontend is our initial setup and will be our
reference for all measurements using `h1load`.

This frontend checks the request hostname using an ACL and route the
traffic to the specific backend (`vhost` in our example) if this ACL
is valid and to `default` otherwise.

The proxy has to be started using:

```
NO_PATTERN_CACHE=1 haproxy -f haproxy.cfg
```

We need to test different cases:

* hostname set to `vhost00.example.com` to see what happens when the
  first rule matches.
* hostname set to `vhost42.example.com` to see what happens when a
  random rule matches.
* hostname set to `vhost100.example.com` to see what happens when a
  no rule matches.

The `h1load` command line is:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S -H 'host: vhost100.example.com' 'http://198.18.0.102:8000/'
```

Change the host header to test different cases.

The following graph shows the requests per second and the latency when
the number tests increases:

![](results-plot.png)

In the worst case scenario (hostname set to `vhost100.example.com`)
the proxy can handle 44k re/s with a 1.4AMS latency.


## First optimization: using anonymous ACL

Maybe we can get rid of all these ACL and do the hostname check in the
`if` condition of each `use_backend` rules. This is how
`anonymous-acl` frontend is set.

Since we know that if a hostname is not matched is our worst case
scenario we will only benchmark for hostname set to
`vhost100.example.com`.

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S -H 'host: vhost100.example.com' 'http://198.18.0.102:8001/'
```

The results are roughly the same as for `origin` frontend. This is
because an ACL is only executed when called. The frontend is a bit
easier to maintain but using an anonymous ACL does not change
anything.


## Second optimization: using a transaction variable

The `req.hdr(host)` has a cost and using it for all rules increases
its cost. We can store the hostname in a variable to reduce its
extraction time. This is how `anonymous-acl-var` frontend is
configured. Now a transaction variable is compared to a known
hostname.

We can benchmark using following command:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S -H 'host: vhost100.example.com' 'http://198.18.0.102:8002/'
```

the results are way better we can reach 68k req/s.

## Third optimization: using a map file

The `backend-map-default` looks up for the hostname into the
`backends.map` map file. The first column if this file is the hostname
and the second one is the backend to use.

Use this command to benchmark the frontend:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S -H 'host: vhost100.example.com' 'http://198.18.0.102:8004/'
```

Now the results are excellent: 150k req/s.

As a bonus the port is trimmed out and the hostname is normalized for
almost not extra cost. And since the `map` lookup accepts a default
value we do not need the `default_backend` directive anymore.
